from blakbox.globs import pg
from blakbox.util import scale_v2
from blakbox.atom import Atom
from blakbox.app.window import Window
from blakbox.app.camera import Camera
from blakbox.app.resource.image import Image
from blakbox.app.resource.surfmap import SurfMap

# ------------------------------------------------------------ #
class Renderer(Atom):
    class FLAGS:
        SHOW_CAMERA: int = 1 << 0  # flag to display the camera's viewport boundaries.

    def __init__(self, surfmap: SurfMap, camera: Camera) -> None:
        super().__init__(0, 0)
        self.draw_calls = 0
        self._draw_calls = []
        self.camera: Camera = camera
        self.surfmap: SurfMap = surfmap
        self.window: Window = camera.window

    def pre_render(self) -> None:   pass
    def post_render(self) -> None:  pass

    def draw_call(self, image: Image, surfid: int=None, pos: list[int]=[0, 0]) -> None:
        """
        Queues a draw call for rendering.

        :param surface: The image/surface to render.
        :param pos: The world-space position of the surface.
        
        Performs frustum culling to avoid rendering objects outside of the viewport.
        """
        if self.draw_calls + 1 > 4096:  # prevent excessive draw calls.
            return

        # frustum culling
        if ((pos[0] + image.size[0]) - self.window.clip_range[0] < self.camera.pos[0] or 
             pos[0] + self.window.clip_range[0] > self.camera.pos[0] + self.camera.bounds[0]) or \
           ((pos[1] + image.size[1]) - self.window.clip_range[1] < self.camera.pos[1] or 
             pos[1] + self.window.clip_range[1] > self.camera.pos[1] + self.camera.bounds[1]):
            return
        
        if isinstance(surfid, int):
            image.set_colorkey([1, 1, 1])
            image.fill([1, 1, 1])
            self.surfmap.blit(surfid, image, [0, 0])

        self._draw_calls.append([image, pos])
        self.draw_calls += 1

    def render(self) -> None:
        self.target = pg.Surface(self.camera.viewport_size)  # create a surface matching the viewport size.
        self.target.fill(self.window.clear_color)
        self.window.fill()

        self.pre_render()
        for i in range(self.draw_calls):
            image, pos = self._draw_calls.pop(0)
            self.window.blit(image, pos)
        self.draw_calls = 0
        self.post_render()

        if self.get_state(self.FLAGS.SHOW_CAMERA):
            self.window.blitr(self.camera.viewport_rect, color=[255, 255, 255], width=1)
            self.window.blitr(self.camera.center_rect([10, 10]), color=[0, 255, 0], width=1)
        
        # apply camera transformations at the render-target level (no per-object transformations)
        self.target.blit(self.window.display, self.camera.offset)
        self.window.screen.blit(
            pg.transform.scale(self.target, self.window.screen_size),
            [0, 0]
        )
        del self.target
# ------------------------------------------------------------ #



from blakbox.globs import pg
from blakbox.util import add_v2, div_v2, scale_v2, clamp

from blakbox.atom import Atom

# ------------------------------------------------------------ #
class Window(Atom):
    def __init__(
            self,
            screen_size: list[int],
            display_size: list[int],
            clip_range: list[int],
            clear_color: list[int],
            ) -> None:
        super().__init__(0, 0)
        self.zoom: float = 1.0
        self.icon: pg.Surface = None
        self.title: str = "BlakWindow"
        self.display = pg.Surface(display_size)
        self.clip_range: list[int] = clip_range[:]
        self.clear_color: list[int] = clear_color[:]
        self.screen_size: list[int] = screen_size[:]
        self.display_size: list[int] = display_size[:]
        self.screen = pg.display.set_mode(screen_size)

    def mod_icon(self, icon: pg.Surface) -> None:
        if not isinstance(icon, pg.Surface): return
        self.icon = icon
        pg.display.set_icon(self.icon)

    def mod_title(self, title: str) -> None:
        if not isinstance(title, str): return
        self.title = title
        pg.display.set_caption(title)

    def mod_display(self, display_size: list[int]) -> None:
        self.display_size = display_size
        self.display = pg.Surface(display_size)

    def mod_zoom(self, delta: float) -> None:
        self.zoom = clamp(self.zoom + delta, 1, 3.5)
    
    def fill(self) -> None:
        self.screen.fill(self.clear_color)
        self.display.fill(self.clear_color)

    def _rad_clip(self, pos: list[float], offset: list[float], radius: float = 0) -> bool:
        pos = add_v2(pos, offset)
        return ((pos[0] + radius < self.clip_range[0])               or
                (pos[0] + self.clip_range[0] > self.display_size[0]) or
                (pos[1] + radius < self.clip_range[1])               or
                (pos[0] + self.clip_range[1] > self.display_size[1]))

    def _blit_clip(self, pos: list[float], size: list[int], offset: list[float]) -> bool:
        pos = add_v2(add_v2(pos, div_v2(size, 2)), offset)
        return ((pos[0] - self.clip_range[0]) < 0 or pos[0] + self.clip_range[0] > self.display_size[0]\
             or (pos[1] - self.clip_range[1]) < 0 or pos[1] + self.clip_range[1] > self.display_size[1])

    def draw_line(
            self, 
            start: list[float],
            end: list[float],
            offset: list[float]=[0, 0],
            color: list[float]=[255, 255, 255], width: int=1) -> None:
        if self._rad_clip(start, offset, 0) or self._rad_clip(end, offset, 0): return
        pg.draw.line(self.display, color, add_v2(start, offset), add_v2(end, offset), width=width)

    def draw_rect(
        self, 
        pos: list[float],
        size: list[float],
        offset: list[float]=[0, 0],
        color: list[float]=[255, 255, 255], width: int=1) -> None:
        if self._blit_clip(pos, size, offset): return
        pg.draw.rect(self.display, color, pg.Rect(add_v2(pos, offset), size), width=width)

    def draw_circle(
        self,
        radius: float,
        center: list[int],
        offset: list[float]=[0, 0],
        color: list[int]=[255, 255, 255], width: int=1) -> None:
        if self._rad_clip(center, offset, radius): return
        pg.draw.circle(self.display, color, add_v2(center, offset), radius, width)

    def blitr(self, rect: pg.Rect, offset: list[float]=[0, 0], color: list[int]=[255, 255, 255], width: int=1) -> None:
        self.draw_rect(rect.topleft, rect.size, offset, color, width)

    def blit(self, surface: pg.Surface, pos: list[int], offset: list[int]=[0, 0], rect: pg.Rect=None) -> None:
        if self._blit_clip(pos, surface.size, offset): return
        self.display.blit(surface, add_v2(pos, offset), rect)

    def update(self) -> None:
        pg.display.flip()
# ------------------------------------------------------------ #



from blakbox.globs import pg
from blakbox.util import abs_path, add_v2, div_v2, scale_v2

from blakbox.atom import Atom
from blakbox.app.proc import Process
from blakbox.app.scene import Scene
from blakbox.app.clock import Clock
from blakbox.app.window import Window
from blakbox.app.camera import Camera
from blakbox.app.renderer import Renderer
from blakbox.app.event import EventManager
from blakbox.app.input import Keyboard, Mouse
from blakbox.app.resource.image import Image
from blakbox.app.resource.surfmap import SurfMap
from blakbox.app.resource.manager import ResourceManager

# ------------------------------------------------------------ #
class Application:
    def __init__(self, name: str="My App", window_size: list[int]=[800, 600]) -> None:
        self.state: int = 0
        self.name:str = name
        self.clock: Clock = Clock()
        self.events: EventManager = EventManager()
        self.surfmap: SurfMap = SurfMap(window_size)
        self.resource: ResourceManager = ResourceManager()

        self.scene: str = None
        self.scene: Scene = None
        self.scenes: dict[str, Scene] = {}

        self.window: Window = Window(window_size, window_size, [1, 1], [58, 94, 123])
        self.window.mod_title(name)
        self.window.mod_icon(pg.image.load(abs_path("assets/images/logo.ico")))

        self.camera: Camera = Camera(self.window)
        self.renderer: Renderer = Renderer(self.surfmap, self.camera)

        self.processes: list[Process] = [None for _ in range(100)]
        self.configure()
    
    def _bsort(self) -> None:
        for _ in range(len(self.processes) - 1, 0, -1):
            for __ in range(_):
                if not self.processes[__] or not self.processes[__+1]: continue
                if self.processes[__].id > self.processes[__+1].id:
                    t = self.processes[__]
                    self.processes[__] = self.processes[__+1]
                    self.processes[__+1] = t
        
    def add_proc(self, proc: Process) -> None:
        if not isinstance(proc, Process): return
        self.processes[proc.id] = proc

    def rem_proc(self, pid: int) -> None:
        try:
            self.processes.pop(pid)
        except IndexError: pass

    def set_scene(self, scene: Scene) -> None:
        self.scenes[scene.name] = scene
        self.scene = self.scenes[scene.name]
    
    def rem_scene(self, key: str) -> Scene|None:
        if self.get_scene(key) is not None:
            self.scenes.pop(key, None)
            self.scene = None
    
    def get_scene(self, key: str) -> Scene|None:
        return self.scenes.get(key, None)

    def configure(self) -> None: raise NotImplementedError
    def handle_events(self) -> None: raise NotImplementedError
    def handle_update(self) -> None: raise NotImplementedError
    def handle_render(self) -> None: raise NotImplementedError

    def exit(self) -> None: raise NotImplementedError
    def run(self) -> None:
        while not self.events.quit:
            self.clock.tick()
            self.events.update()

            if isinstance(self.scene, Scene):
                self.scene.handle_events()
                self.handle_events()

                self.scene.handle_update()
                self.scene.interface.update(self.events)
                self.handle_update()
                self.camera.update(self.clock.delta)

                self.scene.handle_render()
                self.handle_render()
                self.renderer.render()
                self.scene.interface.render()
                
                if isinstance(Mouse.cursor.image, Image):
                    self.window.window.blit(Mouse.cursor.image.data, Mouse.pos.screen)

                Mouse.pos.world = scale_v2(div_v2(Mouse.pos.view, self.scene.tilemap.tilesize), self.scene.tilemap.tilesize)
            else:
                self.handle_events()

                self.handle_update()
                self.camera.update(self.clock.delta)

                self.handle_render()
                self.renderer.render()
                if isinstance(Mouse.cursor.image, Image):
                    self.window.window.blit(Mouse.cursor.image.data, Mouse.pos.screen)

            Mouse.pos.rel = pg.mouse.get_rel()
            Mouse.pos.screen = pg.mouse.get_pos()
            Mouse.pos.view = add_v2([Mouse.pos.screen[0] // self.camera.viewport_scale[0],
                                     Mouse.pos.screen[1] // self.camera.viewport_scale[1]], self.camera.pos)
            self.window.update()
            self.clock.rest()
        else:
            self.exit()
# ------------------------------------------------------------ #



from blakbox.app.clock import Clock
from blakbox.app.proc import Process
from blakbox.app.window import Window
from blakbox.app.camera import Camera
from blakbox.app.renderer import Renderer
from blakbox.app.event import EventManager
from blakbox.app.input import Keyboard, Mouse

from blakbox.app.scene import Scene
from blakbox.app.base import Application

import blakbox.app.ui as ui
import blakbox.app.resource as resource


import blakbox.app as app
import blakbox.game as game
import blakbox.util as util
import blakbox.version as version
import blakbox.quotes as quotes

from blakbox.globs import os, random
if "BLAKBOX_NO_PROMT" not in os.environ:
    print(
        f'BLAKBOX {version.BLAKBOX_YEAR}.{version.BLAKBOX_MINOR}.{version.BLAKBOX_PATCH} | "{random.choice(quotes.BLAKBOX_QUOTES)}"'
    )


from blakbox.globs import pg
from blakbox.util import mag_v2, div_v2, sub_v2, clamp, add_v2

from blakbox.atom import Atom

# ------------------------------------------------------------ #
class Object(Atom):
    def __init__(
            self,
            pos: list[int],
            size: list[int],
            color: list[int],
            bounds: list[int]=None,
    ) -> None:
        super().__init__(0, 0)
        self.size = size[:]
        self.color = color[:]
        self.image = pg.Surface(size)
        
        self.speed = 100
        self.pos = pos[:]
        self.movement = [0, 0, 0, 0, None]
        self.bounds = bounds[:] if bounds else None
        self.fill()
    
    @property
    def center(self) -> list[float]:
        return add_v2(self.pos, div_v2(self.size, 2))
    
    def fill(self) -> None:
        self.image.fill(self.color)
    
    def mod_colorkey(self, color: list[int]) -> None:
        self.image.set_colorkey(color)
    
    def rect(self) -> pg.Rect:
        return pg.Rect(self.pos, self.size)
    
    def move_to(self, pos: list[float]) -> None:
        self.movement[4] = pos[:] if pos else None

    def move(self, l: bool=None, r: bool=None, u: bool=None, d: bool=None) -> None:
        self.movement[0] = True if l else False
        self.movement[1] = True if r else False
        self.movement[2] = True if u else False
        self.movement[3] = True if d else False
    
    def _aabbx(self, transform: list[float], collidables: list["Object"]) -> None:
        rect = self.rect()
        for obj in collidables:
            if rect.colliderect(obj.rect()):
                if transform[0] < 0:
                    rect.left = obj.rect().right
                    self.movement[0] = 0
                if transform[0] > 0:
                    rect.right = obj.rect().left
                    self.movement[1] = 0
                self.pos[0] = rect.x
    
    def _aabby(self, transform: list[float], collidables: list["Object"]) -> None:
        rect = self.rect()
        for obj in collidables:
            if rect.colliderect(obj.rect()):
                if transform[1] < 0:
                    rect.top = obj.rect().bottom
                    self.movement[2] = 0
                if transform[1] > 0:
                    rect.bottom = obj.rect().top
                    self.movement[3] = 0
                self.pos[1] = rect.y

    def update(self, dt: float, collidables: list["Object"]=[]) -> None:
        if self.movement[4]:
            diff = sub_v2(self.movement[4], self.center)
            dist = mag_v2(diff)
            if int(dist) <= 8:
                self.movement[4] = None
            else:
                dirt = div_v2(diff, dist)
                if self.bounds:
                    self.pos[0] = clamp(self.pos[0] + dirt[0] * self.speed * dt, 0, self.bounds[0])
                    self._aabbx(dirt, collidables)
                    self.pos[1] = clamp(self.pos[1] + dirt[1] * self.speed * dt, 0, self.bounds[1])
                    self._aabby(dirt, collidables)
                else:
                    self.pos[0] += dirt[0] * self.speed * dt
                    self._aabbx(dirt, collidables)
                    self.pos[1] += dirt[1] * self.speed * dt
                    self._aabby(dirt, collidables)
        else:        
            t = [(self.movement[1] - self.movement[0]) * self.speed,
                 (self.movement[3] - self.movement[2]) * self.speed]
            if self.bounds:
                self.pos[0] = clamp(self.pos[0] + t[0] * dt, 0, self.bounds[0])
                self._aabbx(t, collidables)
                self.pos[1] = clamp(self.pos[1] + t[1] * dt, 0, self.bounds[1])
                self._aabby(t, collidables)
            else:
                self.pos[0] += t[0] * dt
                self._aabbx(t, collidables)
                self.pos[1] += t[1] * dt
                self._aabby(t, collidables)
# ------------------------------------------------------------ #


from blakbox.app.resource.base import Resource
from blakbox.app.resource.image import Image
from blakbox.app.resource.surfmap import SurfMap
from blakbox.app.resource.animation import Animation
from blakbox.app.resource.manager import ResourceManager


from blakbox.util import add_v2
from blakbox.globs import os, pg
from blakbox.game.obj import Object
from blakbox.app.window import Window

# ------------------------------------------------------------ #
class SurfMap(Object):
    def __init__(self, size: list[int]) -> None:
        super().__init__([0, 0], size, [1, 1, 1])
        self.mod_colorkey([1, 1, 1])
        self.fill()
        self.surf_data: list[tuple[list[int], list[int]]] = []

    def load_surface(self, size: list[int], path: str) -> int:
        if not os.path.exists(path): return
        surfid = len(self.surf_data)

        surface = pg.image.load(path).convert_alpha()
        w, h = size
        if self.pos[0] + w > self.size[0]:
            self.pos[0] = 0
            self.pos[1] += h

        if self.pos[1] + h > self.size[1]: return

        self.image.blit(surface, self.pos)
        del surface

        self.surf_data.append([self.pos[:], size[:]])
        self.pos[0] += w  # move right for next texture

        return surfid

    def unload_surface(self, surfid: int) -> None:
        if surfid < 0 or surfid >= len(self.surf_data): return
        pos, size = self.surf_data[surfid]
        surface = pg.Surface(size)
        surface.fill([1, 1, 1])
        self.image.blit(surface, pos, pg.Rect(pos, size))

    def blit(self, surfid: int, surface: pg.Surface, pos: list[int], offset: list[int]=[0, 0]) -> None:
        if surfid < 0 or surfid >= len(self.surf_data): return
        pos = add_v2(pos, offset)
        surface.blit(self.image, pos, pg.Rect(*self.surf_data[surfid]))
# ------------------------------------------------------------ #



from blakbox.globs import pg, time
from blakbox.atom import Atom

# ------------------------------------------------------------ #
class Clock(Atom):
    def __init__(self, fps: float=60):
        super().__init__(0, 0)
        self._ = pg.time.Clock()
        self.data = [time.time(), time.time(), 0, 0, fps]    # t s dt fps tfps
    
    @property
    def fps(self) -> float:     return self.data[3]
    @property
    def start(self) -> float:   return self.data[1]
    @property
    def delta(self) -> float:   return self.data[2]
    @property
    def current(self) -> float: return self.data[0]

    def tick(self) -> None:
        self.data[2] = time.time() - self.data[0]  # update dt
        self.data[0] = time.time()                 # update t
        self.data[3] = self._.get_fps()            # update fps
    
    def rest(self) -> None:
        self._.tick(self.data[4])
# ------------------------------------------------------------ #

